package java.util;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.regex.MatchResult;
import java.text.DecimalFormatSymbols;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.io.StringReader;
import java.nio.file.Files;
import java.nio.file.OpenOption;
import java.nio.file.Path;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.nio.charset.CharsetDecoder;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.File;
import java.nio.charset.UnsupportedCharsetException;
import java.nio.charset.IllegalCharsetNameException;
import java.nio.charset.Charset;
import java.io.InputStreamReader;
import java.io.InputStream;
import java.io.IOException;
import sun.misc.LRUCache;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.nio.CharBuffer;
import java.io.Closeable;

public final class Scanner implements Iterator<String>, Closeable
{
    private CharBuffer buf;
    private static final int BUFFER_SIZE = 1024;
    private int position;
    private Matcher matcher;
    private Pattern delimPattern;
    private Pattern hasNextPattern;
    private int hasNextPosition;
    private String hasNextResult;
    private Readable source;
    private boolean sourceClosed;
    private boolean needInput;
    private boolean skipped;
    private int savedScannerPosition;
    private Object typeCache;
    private boolean matchValid;
    private boolean closed;
    private int radix;
    private int defaultRadix;
    private Locale locale;
    private LRUCache<String, Pattern> patternCache;
    private IOException lastException;
    private static Pattern WHITESPACE_PATTERN;
    private static Pattern FIND_ANY_PATTERN;
    private static Pattern NON_ASCII_DIGIT;
    private String groupSeparator;
    private String decimalSeparator;
    private String nanString;
    private String infinityString;
    private String positivePrefix;
    private String negativePrefix;
    private String positiveSuffix;
    private String negativeSuffix;
    private static volatile Pattern boolPattern;
    private static final String BOOLEAN_PATTERN = "true|false";
    private Pattern integerPattern;
    private String digits;
    private String non0Digit;
    private int SIMPLE_GROUP_INDEX;
    private static volatile Pattern separatorPattern;
    private static volatile Pattern linePattern;
    private static final String LINE_SEPARATOR_PATTERN = "\r\n|[\n\r\u2028\u2029\u0085]";
    private static final String LINE_PATTERN = ".*(\r\n|[\n\r\u2028\u2029\u0085])|.+$";
    private Pattern floatPattern;
    private Pattern decimalPattern;
    
    private static Pattern boolPattern() {
        Pattern boolPattern = Scanner.boolPattern;
        if (boolPattern == null) {
            boolPattern = (Scanner.boolPattern = Pattern.compile("true|false", 2));
            return boolPattern;
        }
        return boolPattern;
    }
    
    private String buildIntegerPatternString() {
        String string = "((?i)[" + this.digits.substring(0, this.radix) + "\\p{javaDigit}])";
        String string2 = "((" + string + "++)|" + ("(" + this.non0Digit + string + "?" + string + "?(" + this.groupSeparator + string + string + string + ")+)") + ")";
        return "(" + ("([-+]?(" + string2 + "))") + ")|(" + (this.positivePrefix + string2 + this.positiveSuffix) + ")|(" + (this.negativePrefix + string2 + this.negativeSuffix) + ")";
    }
    
    private Pattern integerPattern() {
        if (this.integerPattern == null) {
            this.integerPattern = (Pattern)this.patternCache.forName(this.buildIntegerPatternString());
        }
        return this.integerPattern;
    }
    
    private static Pattern separatorPattern() {
        Pattern separatorPattern = Scanner.separatorPattern;
        if (separatorPattern == null) {
            separatorPattern = (Scanner.separatorPattern = Pattern.compile("\r\n|[\n\r\u2028\u2029\u0085]"));
            return separatorPattern;
        }
        return separatorPattern;
    }
    
    private static Pattern linePattern() {
        Pattern linePattern = Scanner.linePattern;
        if (linePattern == null) {
            linePattern = (Scanner.linePattern = Pattern.compile(".*(\r\n|[\n\r\u2028\u2029\u0085])|.+$"));
            return linePattern;
        }
        return linePattern;
    }
    
    private void buildFloatAndDecimalPattern() {
        String str = "(([0-9\\p{javaDigit}]))";
        String string = "([eE][+-]?" + str + "+)?";
        String string2 = "((" + str + "++)|" + ("(" + this.non0Digit + str + "?" + str + "?(" + this.groupSeparator + str + str + str + ")+)") + ")";
        String string3 = "(" + string2 + "|" + string2 + this.decimalSeparator + str + "*+|" + this.decimalSeparator + str + "++)";
        String string4 = "(NaN|" + this.nanString + "|Infinity|" + this.infinityString + ")";
        String string5 = "(([-+]?" + string3 + string + ")|" + ("(" + this.positivePrefix + string3 + this.positiveSuffix + string + ")") + "|" + ("(" + this.negativePrefix + string3 + this.negativeSuffix + string + ")") + ")";
        this.floatPattern = Pattern.compile(string5 + "|" + "[-+]?0[xX][0-9a-fA-F]*\\.[0-9a-fA-F]+([pP][-+]?[0-9]+)?" + "|" + ("(([-+]?" + string4 + ")|" + ("(" + this.positivePrefix + string4 + this.positiveSuffix + ")") + "|" + ("(" + this.negativePrefix + string4 + this.negativeSuffix + ")") + ")"));
        this.decimalPattern = Pattern.compile(string5);
    }
    
    private Pattern floatPattern() {
        if (this.floatPattern == null) {
            this.buildFloatAndDecimalPattern();
        }
        return this.floatPattern;
    }
    
    private Pattern decimalPattern() {
        if (this.decimalPattern == null) {
            this.buildFloatAndDecimalPattern();
        }
        return this.decimalPattern;
    }
    
    private Scanner(Readable source, Pattern delimPattern) {
        this.sourceClosed = false;
        this.needInput = false;
        this.skipped = false;
        this.savedScannerPosition = -1;
        this.typeCache = null;
        this.matchValid = false;
        this.closed = false;
        this.radix = 10;
        this.defaultRadix = 10;
        this.locale = null;
        this.patternCache = new LRUCache<String, Pattern>(7) {
            protected Pattern create(String regex) {
                return Pattern.compile(regex);
            }
            
            protected boolean hasName(Pattern pattern, String anObject) {
                return pattern.pattern().equals(anObject);
            }
        };
        this.groupSeparator = "\\,";
        this.decimalSeparator = "\\.";
        this.nanString = "NaN";
        this.infinityString = "Infinity";
        this.positivePrefix = "";
        this.negativePrefix = "\\-";
        this.positiveSuffix = "";
        this.negativeSuffix = "";
        this.digits = "0123456789abcdefghijklmnopqrstuvwxyz";
        this.non0Digit = "[\\p{javaDigit}&&[^0]]";
        this.SIMPLE_GROUP_INDEX = 5;
        assert source != null : "source should not be null";
        assert delimPattern != null : "pattern should not be null";
        this.source = source;
        this.delimPattern = delimPattern;
        (this.buf = CharBuffer.allocate(1024)).limit(0);
        (this.matcher = this.delimPattern.matcher(this.buf)).useTransparentBounds(true);
        this.matcher.useAnchoringBounds(false);
        this.useLocale(Locale.getDefault(Locale.Category.FORMAT));
        return;
    }
    
    public Scanner(Readable readable) {
        this((Readable)Objects.requireNonNull(readable, "source"), Scanner.WHITESPACE_PATTERN);
    }
    
    public Scanner(InputStream in) {
        this(new InputStreamReader(in), Scanner.WHITESPACE_PATTERN);
    }
    
    public Scanner(InputStream inputStream, String s) {
        this(makeReadable((InputStream)Objects.requireNonNull(inputStream, "source"), toCharset(s)), Scanner.WHITESPACE_PATTERN);
    }
    
    private static Charset toCharset(String charsetName) {
        Objects.requireNonNull(charsetName, "charsetName");
        try {
            return Charset.forName(charsetName);
        }
        catch (IllegalCharsetNameException | UnsupportedCharsetException cause) {
            throw new IllegalArgumentException(cause);
        }
    }
    
    private static Readable makeReadable(InputStream in, Charset cs) {
        return new InputStreamReader(in, cs);
    }
    
    public Scanner(File file) throws FileNotFoundException {
        this(new FileInputStream(file).getChannel());
    }
    
    public Scanner(File file, String s) throws FileNotFoundException {
        this((File)Objects.requireNonNull(file), toDecoder(s));
    }
    
    private Scanner(File file, CharsetDecoder charsetDecoder) throws FileNotFoundException {
        this(makeReadable(new FileInputStream(file).getChannel(), charsetDecoder));
    }
    
    private static CharsetDecoder toDecoder(String s) {
        Objects.requireNonNull(s, "charsetName");
        try {
            return Charset.forName(s).newDecoder();
        }
        catch (IllegalCharsetNameException | UnsupportedCharsetException ex) {
            throw new IllegalArgumentException(s);
        }
    }
    
    private static Readable makeReadable(ReadableByteChannel ch, CharsetDecoder dec) {
        return Channels.newReader(ch, dec, -1);
    }
    
    public Scanner(Path path) throws IOException {
        this(Files.newInputStream(path, new OpenOption[0]));
    }
    
    public Scanner(Path path, String s) throws IOException {
        this((Path)Objects.requireNonNull(path), toCharset(s));
    }
    
    private Scanner(Path path, Charset charset) throws IOException {
        this(makeReadable(Files.newInputStream(path, new OpenOption[0]), charset));
    }
    
    public Scanner(String s) {
        this(new StringReader(s), Scanner.WHITESPACE_PATTERN);
    }
    
    public Scanner(ReadableByteChannel readableByteChannel) {
        this(makeReadable((ReadableByteChannel)Objects.requireNonNull(readableByteChannel, "source")), Scanner.WHITESPACE_PATTERN);
    }
    
    private static Readable makeReadable(ReadableByteChannel readableByteChannel) {
        return makeReadable(readableByteChannel, Charset.defaultCharset().newDecoder());
    }
    
    public Scanner(ReadableByteChannel readableByteChannel, String s) {
        this(makeReadable((ReadableByteChannel)Objects.requireNonNull(readableByteChannel, "source"), toDecoder(s)), Scanner.WHITESPACE_PATTERN);
    }
    
    private void saveState() {
        this.savedScannerPosition = this.position;
    }
    
    private void revertState() {
        this.position = this.savedScannerPosition;
        this.savedScannerPosition = -1;
        this.skipped = false;
    }
    
    private boolean revertState(boolean b) {
        this.position = this.savedScannerPosition;
        this.savedScannerPosition = -1;
        this.skipped = false;
        return b;
    }
    
    private void cacheResult() {
        this.hasNextResult = this.matcher.group();
        this.hasNextPosition = this.matcher.end();
        this.hasNextPattern = this.matcher.pattern();
    }
    
    private void cacheResult(String hasNextResult) {
        this.hasNextResult = hasNextResult;
        this.hasNextPosition = this.matcher.end();
        this.hasNextPattern = this.matcher.pattern();
    }
    
    private void clearCaches() {
        this.hasNextPattern = null;
        this.typeCache = null;
    }
    
    private String getCachedResult() {
        this.position = this.hasNextPosition;
        this.hasNextPattern = null;
        this.typeCache = null;
        return this.hasNextResult;
    }
    
    private void useTypeCache() {
        if (this.closed) {
            throw new IllegalStateException("Scanner closed");
        }
        this.position = this.hasNextPosition;
        this.hasNextPattern = null;
        this.typeCache = null;
        return;
    }
    
    private void readInput() {
        if (this.buf.limit() == this.buf.capacity()) {
            this.makeSpace();
        }
        int position = this.buf.position();
        this.buf.position(this.buf.limit());
        this.buf.limit(this.buf.capacity());
        int read;
        try {
            read = this.source.read(this.buf);
        }
        catch (IOException lastException) {
            this.lastException = lastException;
            read = -1;
        }
        if (read == -1) {
            this.sourceClosed = true;
            this.needInput = false;
        }
        if (read > 0) {
            this.needInput = false;
        }
        this.buf.limit(this.buf.position());
        this.buf.position(position);
    }
    
    private boolean makeSpace() {
        this.clearCaches();
        int n = (this.savedScannerPosition == -1) ? this.position : this.savedScannerPosition;
        this.buf.position(n);
        if (n > 0) {
            this.buf.compact();
            this.translateSavedIndexes(n);
            this.position -= n;
            this.buf.flip();
            return true;
        }
        CharBuffer allocate = CharBuffer.allocate(this.buf.capacity() * 2);
        allocate.put(this.buf);
        allocate.flip();
        this.translateSavedIndexes(n);
        this.position -= n;
        this.buf = allocate;
        this.matcher.reset(this.buf);
        return true;
    }
    
    private void translateSavedIndexes(int n) {
        if (this.savedScannerPosition != -1) {
            this.savedScannerPosition -= n;
        }
    }
    
    private void throwFor() {
        this.skipped = false;
        if (this.sourceClosed && this.position == this.buf.limit()) {
            throw new NoSuchElementException();
        }
        throw new InputMismatchException();
    }
    
    private boolean hasTokenInBuffer() {
        this.matchValid = false;
        this.matcher.usePattern(this.delimPattern);
        this.matcher.region(this.position, this.buf.limit());
        if (this.matcher.lookingAt()) {
            this.position = this.matcher.end();
        }
        return this.position != this.buf.limit();
    }
    
    private String getCompleteTokenInBuffer(Pattern pattern) {
        this.matchValid = false;
        this.matcher.usePattern(this.delimPattern);
        if (!this.skipped) {
            this.matcher.region(this.position, this.buf.limit());
            if (this.matcher.lookingAt()) {
                if (this.matcher.hitEnd() && !this.sourceClosed) {
                    this.needInput = true;
                    return null;
                }
                this.skipped = true;
                this.position = this.matcher.end();
            }
        }
        if (this.position == this.buf.limit()) {
            if (this.sourceClosed) {
                return null;
            }
            this.needInput = true;
            return null;
        }
        else {
            this.matcher.region(this.position, this.buf.limit());
            int n = this.matcher.find();
            if (n && this.matcher.end() == this.position) {
                n = this.matcher.find();
            }
            if (!(n == 0)) {
                if (this.matcher.requireEnd() && !this.sourceClosed) {
                    this.needInput = true;
                    return null;
                }
                int start = this.matcher.start();
                if (pattern == null) {
                    pattern = Scanner.FIND_ANY_PATTERN;
                }
                this.matcher.usePattern(pattern);
                this.matcher.region(this.position, start);
                if (this.matcher.matches()) {
                    String group2 = this.matcher.group();
                    this.position = this.matcher.end();
                    return group2;
                }
                return null;
            }
            else if (this.sourceClosed) {
                if (pattern == null) {
                    pattern = Scanner.FIND_ANY_PATTERN;
                }
                this.matcher.usePattern(pattern);
                this.matcher.region(this.position, this.buf.limit());
                if (this.matcher.matches()) {
                    String group = this.matcher.group();
                    this.position = this.matcher.end();
                    return group;
                }
                return null;
            }
            else {
                this.needInput = true;
                return null;
            }
        }
    }
    
    private String findPatternInBuffer(Pattern newPattern, int n) {
        this.matchValid = false;
        this.matcher.usePattern(newPattern);
        int limit = this.buf.limit();
        int n2 = -1;
        if (n > 0) {
            n2 = this.position + n;
            if (n2 < limit) {
                limit = n2;
            }
        }
        this.matcher.region(this.position, limit);
        if (this.matcher.find()) {
            if (this.matcher.hitEnd() && !this.sourceClosed) {
                if (limit != n2) {
                    this.needInput = true;
                    return null;
                }
                if (limit == n2 && this.matcher.requireEnd()) {
                    this.needInput = true;
                    return null;
                }
            }
            this.position = this.matcher.end();
            return this.matcher.group();
        }
        else {
            if (this.sourceClosed) {
                return null;
            }
            if (n == 0 || limit != n2) {
                this.needInput = true;
            }
            return null;
        }
    }
    
    private String matchPatternInBuffer(Pattern newPattern) {
        this.matchValid = false;
        this.matcher.usePattern(newPattern);
        this.matcher.region(this.position, this.buf.limit());
        if (this.matcher.lookingAt()) {
            if (this.matcher.hitEnd() && !this.sourceClosed) {
                this.needInput = true;
                return null;
            }
            this.position = this.matcher.end();
            return this.matcher.group();
        }
        else {
            if (this.sourceClosed) {
                return null;
            }
            this.needInput = true;
            return null;
        }
    }
    
    private void ensureOpen() {
        if (this.closed) {
            throw new IllegalStateException("Scanner closed");
        }
        return;
    }
    
    @Override
    public void close() {
        if (this.closed) {
            return;
        }
        if ((this.source instanceof Closeable)) {
            try {
                ((Closeable)this.source).close();
            }
            catch (IOException lastException) {
                this.lastException = lastException;
            }
        }
        this.sourceClosed = true;
        this.source = null;
        this.closed = true;
        return;
    }
    
    public IOException ioException() {
        return this.lastException;
    }
    
    public Pattern delimiter() {
        return this.delimPattern;
    }
    
    public Scanner useDelimiter(Pattern delimPattern) {
        this.delimPattern = delimPattern;
        return this;
    }
    
    public Scanner useDelimiter(String s) {
        this.delimPattern = (Pattern)this.patternCache.forName(s);
        return this;
    }
    
    public Locale locale() {
        return this.locale;
    }
    
    public Scanner useLocale(Locale locale) {
        if (locale.equals(this.locale)) {
            return this;
        }
        this.locale = locale;
        DecimalFormat decimalFormat = (DecimalFormat)NumberFormat.getNumberInstance(locale);
        DecimalFormatSymbols instance = DecimalFormatSymbols.getInstance(locale);
        this.groupSeparator = "\\x{" + Integer.toHexString((int)instance.getGroupingSeparator()) + "}";
        this.decimalSeparator = "\\x{" + Integer.toHexString((int)instance.getDecimalSeparator()) + "}";
        this.nanString = Pattern.quote(instance.getNaN());
        this.infinityString = Pattern.quote(instance.getInfinity());
        this.positivePrefix = decimalFormat.getPositivePrefix();
        if (!this.positivePrefix.isEmpty()) {
            this.positivePrefix = Pattern.quote(this.positivePrefix);
        }
        this.negativePrefix = decimalFormat.getNegativePrefix();
        if (!this.negativePrefix.isEmpty()) {
            this.negativePrefix = Pattern.quote(this.negativePrefix);
        }
        this.positiveSuffix = decimalFormat.getPositiveSuffix();
        if (!this.positiveSuffix.isEmpty()) {
            this.positiveSuffix = Pattern.quote(this.positiveSuffix);
        }
        this.negativeSuffix = decimalFormat.getNegativeSuffix();
        if (!this.negativeSuffix.isEmpty()) {
            this.negativeSuffix = Pattern.quote(this.negativeSuffix);
        }
        this.integerPattern = null;
        this.floatPattern = null;
        return this;
    }
    
    public int radix() {
        return this.defaultRadix;
    }
    
    public Scanner useRadix(int n) {
        if (n < 2 || n > 36) {
            throw new IllegalArgumentException("radix:" + n);
        }
        if (this.defaultRadix == n) {
            return this;
        }
        this.defaultRadix = n;
        this.integerPattern = null;
        return this;
    }
    
    private void setRadix(int radix) {
        if (this.radix != radix) {
            this.integerPattern = null;
            this.radix = radix;
        }
    }
    
    public MatchResult match() {
        if (!this.matchValid) {
            throw new IllegalStateException("No match result available");
        }
        return this.matcher.toMatchResult();
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("java.util.Scanner");
        sb.append("[delimiters=" + this.delimPattern + "]");
        sb.append("[position=" + this.position + "]");
        sb.append("[match valid=" + this.matchValid + "]");
        sb.append("[need input=" + this.needInput + "]");
        sb.append("[source closed=" + this.sourceClosed + "]");
        sb.append("[skipped=" + this.skipped + "]");
        sb.append("[group separator=" + this.groupSeparator + "]");
        sb.append("[decimal separator=" + this.decimalSeparator + "]");
        sb.append("[positive prefix=" + this.positivePrefix + "]");
        sb.append("[negative prefix=" + this.negativePrefix + "]");
        sb.append("[positive suffix=" + this.positiveSuffix + "]");
        sb.append("[negative suffix=" + this.negativeSuffix + "]");
        sb.append("[NaN string=" + this.nanString + "]");
        sb.append("[infinity string=" + this.infinityString + "]");
        return sb.toString();
    }
    
    @Override
    public boolean hasNext() {
        this.ensureOpen();
        this.saveState();
        while (!this.sourceClosed) {
            if (this.hasTokenInBuffer()) {
                return this.revertState(true);
            }
            this.readInput();
        }
        return this.revertState(this.hasTokenInBuffer());
    }
    
    @Override
    public String next() {
        this.ensureOpen();
        this.clearCaches();
        String completeTokenInBuffer;
        while (true) {
            completeTokenInBuffer = this.getCompleteTokenInBuffer(null);
            if (completeTokenInBuffer != null) {
                break;
            }
            if (this.needInput) {
                this.readInput();
            }
            else {
                this.throwFor();
            }
        }
        this.matchValid = true;
        this.skipped = false;
        return completeTokenInBuffer;
    }
    
    @Override
    public void remove() {
        throw new UnsupportedOperationException();
    }
    
    public boolean hasNext(String s) {
        return this.hasNext((Pattern)this.patternCache.forName(s));
    }
    
    public String next(String s) {
        return this.next((Pattern)this.patternCache.forName(s));
    }
    
    public boolean hasNext(Pattern pattern) {
        this.ensureOpen();
        if (pattern == null) {
            throw new NullPointerException();
        }
        this.hasNextPattern = null;
        this.saveState();
        while (this.getCompleteTokenInBuffer(pattern) == null) {
            if (this.needInput) {
                this.readInput();
            }
            else {
                return this.revertState(false);
            }
        }
        this.matchValid = true;
        this.cacheResult();
        return this.revertState(true);
    }
    
    public String next(Pattern pattern) {
        this.ensureOpen();
        if (pattern == null) {
            throw new NullPointerException();
        }
        if (this.hasNextPattern == pattern) {
            return this.getCachedResult();
        }
        this.clearCaches();
        String completeTokenInBuffer;
        while (true) {
            completeTokenInBuffer = this.getCompleteTokenInBuffer(pattern);
            if (completeTokenInBuffer != null) {
                break;
            }
            if (this.needInput) {
                this.readInput();
            }
            else {
                this.throwFor();
            }
        }
        this.matchValid = true;
        this.skipped = false;
        return completeTokenInBuffer;
    }
    
    public boolean hasNextLine() {
        this.saveState();
        String s = this.findWithinHorizon(linePattern(), 0);
        if (s != null) {
            String group = this.match().group(1);
            if (group != null) {
                s = s.substring(0, s.length() - group.length());
                this.cacheResult(s);
            }
            else {
                this.cacheResult();
            }
        }
        this.revertState();
        return s != null;
    }
    
    public String nextLine() {
        if (this.hasNextPattern == linePattern()) {
            return this.getCachedResult();
        }
        this.clearCaches();
        String s = this.findWithinHorizon(Scanner.linePattern, 0);
        if (s == null) {
            throw new NoSuchElementException("No line found");
        }
        String group = this.match().group(1);
        if (group != null) {
            s = s.substring(0, s.length() - group.length());
        }
        if (s == null) {
            throw new NoSuchElementException();
        }
        return s;
    }
    
    public String findInLine(String s) {
        return this.findInLine((Pattern)this.patternCache.forName(s));
    }
    
    public String findInLine(Pattern pattern) {
        this.ensureOpen();
        if (pattern == null) {
            throw new NullPointerException();
        }
        this.clearCaches();
        this.saveState();
        Label_0078: {
            while (this.findPatternInBuffer(separatorPattern(), 0) == null) {
                if (this.needInput) {
                    this.readInput();
                }
                else {
                    int n = this.buf.limit();
                    this.revertState();
                    int n2 = n - this.position;
                    if (n2 == 0) {
                        return null;
                    }
                    return this.findWithinHorizon(pattern, n2);
                }
            }
            int n = this.matcher.start();
            continue;
        }
    }
    
    public String findWithinHorizon(String s, int n) {
        return this.findWithinHorizon((Pattern)this.patternCache.forName(s), n);
    }
    
    public String findWithinHorizon(Pattern pattern, int n) {
        this.ensureOpen();
        if (pattern == null) {
            throw new NullPointerException();
        }
        if (n < 0) {
            throw new IllegalArgumentException("horizon < 0");
        }
        this.clearCaches();
        while (true) {
            String patternInBuffer = this.findPatternInBuffer(pattern, n);
            if (patternInBuffer != null) {
                this.matchValid = true;
                return patternInBuffer;
            }
            if (this.needInput) {
                this.readInput();
            }
            else {
                return null;
            }
        }
    }
    
    public Scanner skip(Pattern pattern) {
        this.ensureOpen();
        if (pattern == null) {
            throw new NullPointerException();
        }
        this.clearCaches();
        while (this.matchPatternInBuffer(pattern) == null) {
            if (this.needInput) {
                this.readInput();
            }
            else {
                throw new NoSuchElementException();
            }
        }
        this.matchValid = true;
        this.position = this.matcher.end();
        return this;
    }
    
    public Scanner skip(String s) {
        return this.skip((Pattern)this.patternCache.forName(s));
    }
    
    public boolean hasNextBoolean() {
        return this.hasNext(boolPattern());
    }
    
    public boolean nextBoolean() {
        this.clearCaches();
        return Boolean.parseBoolean(this.next(boolPattern()));
    }
    
    public boolean hasNextByte() {
        return this.hasNextByte(this.defaultRadix);
    }
    
    public boolean hasNextByte(int n) {
        this.setRadix(n);
        int hasNext = this.hasNext(this.integerPattern());
        if (!(hasNext == 0)) {
            try {
                this.typeCache = Byte.valueOf(Byte.parseByte((this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) ? this.processIntegerToken(this.hasNextResult) : this.hasNextResult, n));
            }
            catch (NumberFormatException ex) {
                hasNext = false;
            }
        }
        return hasNext;
    }
    
    public byte nextByte() {
        return this.nextByte(this.defaultRadix);
    }
    
    public byte nextByte(int n) {
        if (this.typeCache != null && this.typeCache instanceof Byte && this.radix == n) {
            int byteValue = (byte)this.typeCache;
            this.useTypeCache();
            return byteValue;
        }
        this.setRadix(n);
        this.clearCaches();
        try {
            String s = this.next(this.integerPattern());
            if (this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) {
                s = this.processIntegerToken(s);
            }
            return Byte.parseByte(s, n);
        }
        catch (NumberFormatException ex) {
            this.position = this.matcher.start();
            throw new InputMismatchException(ex.getMessage());
        }
    }
    
    public boolean hasNextShort() {
        return this.hasNextShort(this.defaultRadix);
    }
    
    public boolean hasNextShort(int n) {
        this.setRadix(n);
        int hasNext = this.hasNext(this.integerPattern());
        if (!(hasNext == 0)) {
            try {
                this.typeCache = Short.valueOf(Short.parseShort((this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) ? this.processIntegerToken(this.hasNextResult) : this.hasNextResult, n));
            }
            catch (NumberFormatException ex) {
                hasNext = false;
            }
        }
        return hasNext;
    }
    
    public short nextShort() {
        return this.nextShort(this.defaultRadix);
    }
    
    public short nextShort(int n) {
        if (this.typeCache != null && this.typeCache instanceof Short && this.radix == n) {
            int shortValue = (short)this.typeCache;
            this.useTypeCache();
            return shortValue;
        }
        this.setRadix(n);
        this.clearCaches();
        try {
            String s = this.next(this.integerPattern());
            if (this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) {
                s = this.processIntegerToken(s);
            }
            return Short.parseShort(s, n);
        }
        catch (NumberFormatException ex) {
            this.position = this.matcher.start();
            throw new InputMismatchException(ex.getMessage());
        }
    }
    
    public boolean hasNextInt() {
        return this.hasNextInt(this.defaultRadix);
    }
    
    public boolean hasNextInt(int n) {
        this.setRadix(n);
        int hasNext = this.hasNext(this.integerPattern());
        if (!(hasNext == 0)) {
            try {
                this.typeCache = Integer.valueOf(Integer.parseInt((this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) ? this.processIntegerToken(this.hasNextResult) : this.hasNextResult, n));
            }
            catch (NumberFormatException ex) {
                hasNext = false;
            }
        }
        return hasNext;
    }
    
    private String processIntegerToken(String s) {
        String str = s.replaceAll("" + this.groupSeparator, "");
        boolean n = false;
        int length = this.negativePrefix.length();
        if (length > 0 && str.startsWith(this.negativePrefix)) {
            n = true;
            str = str.substring(length);
        }
        int length2 = this.negativeSuffix.length();
        if (length2 > 0 && str.endsWith(this.negativeSuffix)) {
            n = true;
            str = str.substring(str.length() - length2, str.length());
        }
        if (!(n == 0)) {
            str = "-" + str;
            return str;
        }
        return str;
    }
    
    public int nextInt() {
        return this.nextInt(this.defaultRadix);
    }
    
    public int nextInt(int n) {
        if (this.typeCache != null && this.typeCache instanceof Integer && this.radix == n) {
            int intValue = (int)this.typeCache;
            this.useTypeCache();
            return intValue;
        }
        this.setRadix(n);
        this.clearCaches();
        try {
            String s = this.next(this.integerPattern());
            if (this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) {
                s = this.processIntegerToken(s);
            }
            return Integer.parseInt(s, n);
        }
        catch (NumberFormatException ex) {
            this.position = this.matcher.start();
            throw new InputMismatchException(ex.getMessage());
        }
    }
    
    public boolean hasNextLong() {
        return this.hasNextLong(this.defaultRadix);
    }
    
    public boolean hasNextLong(int n) {
        this.setRadix(n);
        int hasNext = this.hasNext(this.integerPattern());
        if (!(hasNext == 0)) {
            try {
                this.typeCache = Long.valueOf(Long.parseLong((this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) ? this.processIntegerToken(this.hasNextResult) : this.hasNextResult, n));
            }
            catch (NumberFormatException ex) {
                hasNext = false;
            }
        }
        return hasNext;
    }
    
    public long nextLong() {
        return this.nextLong(this.defaultRadix);
    }
    
    public long nextLong(int n) {
        if (this.typeCache != null && this.typeCache instanceof Long && this.radix == n) {
            long longValue = (long)this.typeCache;
            this.useTypeCache();
            return longValue;
        }
        this.setRadix(n);
        this.clearCaches();
        try {
            String s = this.next(this.integerPattern());
            if (this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) {
                s = this.processIntegerToken(s);
            }
            return Long.parseLong(s, n);
        }
        catch (NumberFormatException ex) {
            this.position = this.matcher.start();
            throw new InputMismatchException(ex.getMessage());
        }
    }
    
    private String processFloatToken(String s) {
        String s2 = s.replaceAll(this.groupSeparator, "");
        if (!this.decimalSeparator.equals("\\.")) {
            s2 = s2.replaceAll(this.decimalSeparator, ".");
        }
        boolean n = false;
        int length = this.negativePrefix.length();
        if (length > 0 && s2.startsWith(this.negativePrefix)) {
            n = true;
            s2 = s2.substring(length);
        }
        int length2 = this.negativeSuffix.length();
        if (length2 > 0 && s2.endsWith(this.negativeSuffix)) {
            n = true;
            s2 = s2.substring(s2.length() - length2, s2.length());
        }
        if (s2.equals(this.nanString)) {
            s2 = "NaN";
        }
        if (s2.equals(this.infinityString)) {
            s2 = "Infinity";
        }
        if (!(n == 0)) {
            s2 = "-" + s2;
        }
        if (Scanner.NON_ASCII_DIGIT.matcher(s2).find()) {
            StringBuilder o = new StringBuilder();
            for (int i = 0; i < s2.length(); ++i) {
                int char1 = s2.charAt(i);
                if (Character.isDigit((char)char1)) {
                    int digit = Character.digit((char)char1, 10);
                    if (digit != -1) {
                        o.append(digit);
                    }
                    else {
                        o.append((char)char1);
                    }
                }
                else {
                    o.append((char)char1);
                }
            }
            s2 = o.toString();
            return s2;
        }
        return s2;
    }
    
    public boolean hasNextFloat() {
        this.setRadix(10);
        int hasNext = this.hasNext(this.floatPattern());
        if (!(hasNext == 0)) {
            try {
                this.typeCache = Float.valueOf(Float.parseFloat(this.processFloatToken(this.hasNextResult)));
            }
            catch (NumberFormatException ex) {
                hasNext = false;
            }
        }
        return hasNext;
    }
    
    public float nextFloat() {
        if (this.typeCache != null && this.typeCache instanceof Float) {
            float floatValue = (float)this.typeCache;
            this.useTypeCache();
            return floatValue;
        }
        this.setRadix(10);
        this.clearCaches();
        try {
            return Float.parseFloat(this.processFloatToken(this.next(this.floatPattern())));
        }
        catch (NumberFormatException ex) {
            this.position = this.matcher.start();
            throw new InputMismatchException(ex.getMessage());
        }
    }
    
    public boolean hasNextDouble() {
        this.setRadix(10);
        int hasNext = this.hasNext(this.floatPattern());
        if (!(hasNext == 0)) {
            try {
                this.typeCache = Double.valueOf(Double.parseDouble(this.processFloatToken(this.hasNextResult)));
            }
            catch (NumberFormatException ex) {
                hasNext = false;
            }
        }
        return hasNext;
    }
    
    public double nextDouble() {
        if (this.typeCache != null && this.typeCache instanceof Double) {
            double doubleValue = (double)this.typeCache;
            this.useTypeCache();
            return doubleValue;
        }
        this.setRadix(10);
        this.clearCaches();
        try {
            return Double.parseDouble(this.processFloatToken(this.next(this.floatPattern())));
        }
        catch (NumberFormatException ex) {
            this.position = this.matcher.start();
            throw new InputMismatchException(ex.getMessage());
        }
    }
    
    public boolean hasNextBigInteger() {
        return this.hasNextBigInteger(this.defaultRadix);
    }
    
    public boolean hasNextBigInteger(int n) {
        this.setRadix(n);
        int hasNext = this.hasNext(this.integerPattern());
        if (!(hasNext == 0)) {
            try {
                this.typeCache = new BigInteger((this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) ? this.processIntegerToken(this.hasNextResult) : this.hasNextResult, n);
            }
            catch (NumberFormatException ex) {
                hasNext = false;
            }
        }
        return hasNext;
    }
    
    public BigInteger nextBigInteger() {
        return this.nextBigInteger(this.defaultRadix);
    }
    
    public BigInteger nextBigInteger(int n) {
        if (this.typeCache != null && this.typeCache instanceof BigInteger && this.radix == n) {
            BigInteger bigInteger = (BigInteger)this.typeCache;
            this.useTypeCache();
            return bigInteger;
        }
        this.setRadix(n);
        this.clearCaches();
        try {
            String val = this.next(this.integerPattern());
            if (this.matcher.group(this.SIMPLE_GROUP_INDEX) == null) {
                val = this.processIntegerToken(val);
            }
            return new BigInteger(val, n);
        }
        catch (NumberFormatException ex) {
            this.position = this.matcher.start();
            throw new InputMismatchException(ex.getMessage());
        }
    }
    
    public boolean hasNextBigDecimal() {
        this.setRadix(10);
        int hasNext = this.hasNext(this.decimalPattern());
        if (!(hasNext == 0)) {
            try {
                this.typeCache = new BigDecimal(this.processFloatToken(this.hasNextResult));
            }
            catch (NumberFormatException ex) {
                hasNext = false;
            }
        }
        return hasNext;
    }
    
    public BigDecimal nextBigDecimal() {
        if (this.typeCache != null && this.typeCache instanceof BigDecimal) {
            BigDecimal bigDecimal = (BigDecimal)this.typeCache;
            this.useTypeCache();
            return bigDecimal;
        }
        this.setRadix(10);
        this.clearCaches();
        try {
            return new BigDecimal(this.processFloatToken(this.next(this.decimalPattern())));
        }
        catch (NumberFormatException ex) {
            this.position = this.matcher.start();
            throw new InputMismatchException(ex.getMessage());
        }
    }
    
    public Scanner reset() {
        this.delimPattern = Scanner.WHITESPACE_PATTERN;
        this.useLocale(Locale.getDefault(Locale.Category.FORMAT));
        this.useRadix(10);
        this.clearCaches();
        return this;
    }
    
    static {
        Scanner.WHITESPACE_PATTERN = Pattern.compile("\\p{javaWhitespace}+");
        Scanner.FIND_ANY_PATTERN = Pattern.compile("(?s).*");
        Scanner.NON_ASCII_DIGIT = Pattern.compile("[\\p{javaDigit}&&[^0-9]]");
    }
}
